@@ -1,6 +1,7 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
  *
  * Copyright (C) 2012 Red Hat, Inc.
+ * Copyright (C) 2020 Canonical Ltd.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -29,6 +30,7 @@
 #include <glib/gi18n.h>
 #include <glib/gstdio.h>
 #include <gio/gio.h>
+#include <p11-kit/p11-kit.h>
 
 struct _GsdSmartcardService
 {
@@ -95,7 +97,7 @@ set_bus_connection (GsdSmartcardService  *self,
 static void
 register_object_manager (GsdSmartcardService *self)
 {
-        GsdSmartcardServiceObjectSkeleton *object;
+        g_autoptr(GsdSmartcardServiceObjectSkeleton) object = NULL;
 
         self->object_manager_server = g_dbus_object_manager_server_new (GSD_SMARTCARD_DBUS_PATH);
 
@@ -105,8 +107,6 @@ register_object_manager (GsdSmartcardService *self)
 
         g_dbus_object_manager_server_export (self->object_manager_server,
                                              G_DBUS_OBJECT_SKELETON (object));
-        g_object_unref (object);
-
         g_dbus_object_manager_server_set_connection (self->object_manager_server,
                                                      self->bus_connection);
 }
@@ -120,12 +120,12 @@ get_login_token_object_path (GsdSmartcardService *self)
 static void
 register_login_token_alias (GsdSmartcardService *self)
 {
-        GDBusObjectSkeleton *object;
-        GDBusInterfaceSkeleton *interface;
+        g_autoptr(GDBusObjectSkeleton) object = NULL;
+        g_autoptr(GDBusInterfaceSkeleton) interface = NULL;
         const char *object_path;
         const char *token_name;
 
-        token_name = g_getenv ("PKCS11_LOGIN_TOKEN_NAME");
+        token_name = gsd_smartcard_utils_get_login_token_name ();
 
         if (token_name == NULL)
                 return;
@@ -135,7 +135,6 @@ register_login_token_alias (GsdSmartcardService *self)
         interface = G_DBUS_INTERFACE_SKELETON (gsd_smartcard_service_token_skeleton_new ());
 
         g_dbus_object_skeleton_add_interface (object, interface);
-        g_object_unref (interface);
 
         g_object_set (G_OBJECT (interface),
                       "name", token_name,
@@ -154,8 +153,9 @@ register_login_token_alias (GsdSmartcardService *self)
 static void
 on_bus_gotten (GObject      *source_object,
                GAsyncResult *result,
-               GTask        *task)
+               gpointer      user_data)
 {
+        g_autoptr(GTask) task = g_steal_pointer (&user_data);
         GsdSmartcardService *self;
         GDBusConnection *connection;
         GError *error = NULL;
@@ -163,7 +163,7 @@ on_bus_gotten (GObject      *source_object,
         connection = g_bus_get_finish (result, &error);
         if (connection == NULL) {
                 g_task_return_error (task, error);
-                goto out;
+                return;
         }
 
         g_debug ("taking name %s on session bus", GSD_SMARTCARD_DBUS_NAME);
@@ -186,10 +186,6 @@ on_bus_gotten (GObject      *source_object,
          */
         register_login_token_alias (self);
         g_task_return_boolean (task, TRUE);
-
-out:
-        g_object_unref (task);
-        return;
 }
 
 static gboolean
@@ -228,26 +224,27 @@ async_initable_interface_init (GAsyncInitableIface *interface)
 }
 
 static char *
-get_object_path_for_token (GsdSmartcardService *self,
-                           PK11SlotInfo        *card_slot)
+module_get_path (GckModule *module)
 {
-        char *object_path;
-        char *escaped_library_path;
-        SECMODModule *driver;
-        CK_SLOT_ID slot_id;
+        return p11_kit_config_option (gck_module_get_functions (module), "module");
+}
 
-        driver = PK11_GetModule (card_slot);
-        slot_id = PK11_GetSlotID (card_slot);
+static char *
+get_object_path_for_token (GsdSmartcardService *self,
+                           GckSlot             *slot)
+{
+        g_autofree char *module_path = NULL;
+        g_autofree char *escaped_library_path = NULL;
+        g_autoptr(GckModule) module = NULL;
 
-        escaped_library_path = gsd_smartcard_utils_escape_object_path (driver->dllName);
+        module = gck_slot_get_module (slot);
+        module_path = module_get_path (module);
+        escaped_library_path = gsd_smartcard_utils_escape_object_path (module_path);
 
-        object_path = g_strdup_printf ("%s/token_from_%s_slot_%lu",
-                                       GSD_SMARTCARD_MANAGER_TOKENS_DBUS_PATH,
-                                       escaped_library_path,
-                                       (gulong) slot_id);
-        g_free (escaped_library_path);
-
-        return object_path;
+        return g_strdup_printf ("%s/token_from_%s_slot_%lu",
+                                GSD_SMARTCARD_MANAGER_TOKENS_DBUS_PATH,
+                                escaped_library_path,
+                                gck_slot_get_handle (slot));
 }
 
 static gboolean
@@ -255,8 +252,8 @@ gsd_smartcard_service_handle_get_login_token (GsdSmartcardServiceManager *manage
                                               GDBusMethodInvocation      *invocation)
 {
         GsdSmartcardService *self = GSD_SMARTCARD_SERVICE (manager);
-        PK11SlotInfo *card_slot;
-        char *object_path;
+        g_autoptr(GckSlot) card_slot = NULL;
+        g_autofree char *object_path = NULL;
 
         card_slot = gsd_smartcard_manager_get_login_token (self->smartcard_manager);
 
@@ -288,8 +285,6 @@ gsd_smartcard_service_handle_get_login_token (GsdSmartcardServiceManager *manage
         gsd_smartcard_service_manager_complete_get_login_token (manager,
                                                                 invocation,
                                                                 object_path);
-        g_free (object_path);
-
         return TRUE;
 }
 
@@ -298,29 +293,27 @@ gsd_smartcard_service_handle_get_inserted_tokens (GsdSmartcardServiceManager *ma
                                                   GDBusMethodInvocation      *invocation)
 {
         GsdSmartcardService *self = GSD_SMARTCARD_SERVICE (manager);
-        GList *inserted_tokens, *node;
-        GPtrArray *object_paths;
+        g_autolist(GckSlot) inserted_tokens = NULL;
+        g_autoptr(GPtrArray) object_paths = NULL;
+        GList *node;
 
         inserted_tokens = gsd_smartcard_manager_get_inserted_tokens (self->smartcard_manager,
                                                                      NULL);
 
-        object_paths = g_ptr_array_new ();
+        object_paths = g_ptr_array_new_with_free_func (g_free);
         for (node = inserted_tokens; node != NULL; node = node->next) {
-                PK11SlotInfo *card_slot = node->data;
+                GckSlot *card_slot = node->data;
                 char *object_path;
 
                 object_path = get_object_path_for_token (self, card_slot);
                 g_ptr_array_add (object_paths, object_path);
         }
         g_ptr_array_add (object_paths, NULL);
-        g_list_free (inserted_tokens);
 
         gsd_smartcard_service_manager_complete_get_inserted_tokens (manager,
                                                                     invocation,
                                                                     (const char * const *) object_paths->pdata);
 
-        g_ptr_array_free (object_paths, TRUE);
-
         return TRUE;
 }
 
@@ -345,6 +338,8 @@ gsd_smartcard_service_dispose (GObject *object)
 {
         GsdSmartcardService *self = GSD_SMARTCARD_SERVICE (object);
 
+        g_clear_handle_id (&self->name_id, g_bus_unown_name);
+
         g_clear_object (&self->bus_connection);
         g_clear_object (&self->object_manager_server);
         g_clear_object (&self->smartcard_manager);
@@ -422,8 +417,9 @@ gsd_smartcard_service_class_init (GsdSmartcardServiceClass *service_class)
 static void
 on_new_async_finished (GObject      *source_object,
                        GAsyncResult *result,
-                       GTask        *task)
+                       gpointer      user_data)
 {
+        g_autoptr(GTask) task = g_steal_pointer (&user_data);
         GError *error = NULL;
         GObject *object;
 
@@ -433,15 +429,12 @@ on_new_async_finished (GObject      *source_object,
 
         if (object == NULL) {
                 g_task_return_error (task, error);
-                goto out;
+                return;
         }
 
         g_assert (GSD_IS_SMARTCARD_SERVICE (object));
 
         g_task_return_pointer (task, object, g_object_unref);
-out:
-        g_object_unref (task);
-        return;
 }
 
 void
@@ -481,70 +474,93 @@ gsd_smartcard_service_new_finish (GAsyncResult  *result,
 }
 
 static char *
-get_object_path_for_driver (GsdSmartcardService *self,
-                            SECMODModule        *driver)
+get_object_path_for_module_path (GsdSmartcardService *self,
+                                 const char          *module_path)
 {
         char *object_path;
-        char *escaped_library_path;
+        g_autofree char *escaped_library_path = NULL;
 
-        escaped_library_path = gsd_smartcard_utils_escape_object_path (driver->dllName);
+        escaped_library_path = gsd_smartcard_utils_escape_object_path (module_path);
 
         object_path = g_build_path ("/",
                                     GSD_SMARTCARD_MANAGER_DRIVERS_DBUS_PATH,
                                     escaped_library_path, NULL);
-        g_free (escaped_library_path);
 
         return object_path;
 }
 
-void
-gsd_smartcard_service_register_driver (GsdSmartcardService  *self,
-                                       SECMODModule         *driver)
+static char *
+get_object_path_for_module (GsdSmartcardService *self,
+                            GckModule           *module)
 {
-        char *object_path;
-        GDBusObjectSkeleton *object;
-        GDBusInterfaceSkeleton *interface;
+        g_autofree char *module_path = NULL;
 
-        object_path = get_object_path_for_driver (self, driver);
+        module_path = module_get_path (module);
+
+        return get_object_path_for_module_path (self, module_path);
+}
+
+void
+gsd_smartcard_service_register_driver (GsdSmartcardService *self,
+                                       GckModule           *module)
+{
+        g_autoptr(GDBusObjectSkeleton) object = NULL;
+        g_autoptr(GDBusInterfaceSkeleton) interface = NULL;
+        g_autoptr(GckModuleInfo) module_info = NULL;
+        g_autofree char *object_path = NULL;
+        g_autofree char *module_path = NULL;
+        const char *module_description;
+
+        module_path = module_get_path (module);
+        object_path = get_object_path_for_module_path (self, module_path);
         object = G_DBUS_OBJECT_SKELETON (gsd_smartcard_service_object_skeleton_new (object_path));
-        g_free (object_path);
 
         interface = G_DBUS_INTERFACE_SKELETON (gsd_smartcard_service_driver_skeleton_new ());
         g_dbus_object_skeleton_add_interface (object, interface);
-        g_object_unref (interface);
+
+        module_info = gck_module_get_info (module);
+        module_description = module_info ? module_info->library_description : NULL;
+
+        g_debug ("Registering driver %s", module_description);
 
         g_object_set (G_OBJECT (interface),
-                      "library", driver->dllName,
-                      "description", driver->commonName,
+                      "library", module_path,
+                      "description", module_description,
                       NULL);
         g_dbus_object_manager_server_export (self->object_manager_server,
                                              object);
-        g_object_unref (object);
 }
 
 static void
 synchronize_token_now (GsdSmartcardService *self,
-                       PK11SlotInfo        *card_slot)
+                       GckSlot             *card_slot)
 {
         GDBusInterfaceSkeleton *interface;
-        char *object_path;
-        const char *token_name;
+        g_autoptr(GckTokenInfo) token_info = NULL;
+        g_autoptr(GMutexLocker) locked = NULL;
+        g_autofree char *object_path = NULL;
+        const char *token_name = NULL;
         gboolean is_present, is_login_card;
 
         object_path = get_object_path_for_token (self, card_slot);
 
-        G_LOCK (gsd_smartcard_tokens);
+        locked = g_mutex_locker_new (&G_LOCK_NAME (gsd_smartcard_tokens));
         interface = g_hash_table_lookup (self->tokens, object_path);
-        g_free (object_path);
 
         if (interface == NULL)
-                goto out;
+                return;
 
-        token_name = PK11_GetTokenName (card_slot);
-        is_present = PK11_IsPresent (card_slot);
+        is_present = gck_slot_has_flags (card_slot, CKF_TOKEN_PRESENT);
 
-        if (g_strcmp0 (g_getenv ("PKCS11_LOGIN_TOKEN_NAME"), token_name) == 0)
-                is_login_card = TRUE;
+        if (is_present) {
+                token_info = gck_slot_get_token_info (card_slot);
+
+                if (token_info)
+                        token_name = token_info->label;
+        }
+
+        if (g_strcmp0 (gsd_smartcard_utils_get_login_token_name (), token_name) == 0)
+                is_login_card = is_present;
         else
                 is_login_card = FALSE;
 
@@ -591,14 +607,12 @@ synchronize_token_now (GsdSmartcardService *self,
                         self->login_token_bound = TRUE;
                 }
         }
-
-out:
-        G_UNLOCK (gsd_smartcard_tokens);
 }
 
 typedef struct
 {
-        PK11SlotInfo *card_slot;
+        GckSlot      *card_slot;
+        GckTokenInfo *token_info;
         char         *object_path;
         GSource      *main_thread_source;
 } RegisterNewTokenOperation;
@@ -608,7 +622,8 @@ destroy_register_new_token_operation (RegisterNewTokenOperation *operation)
 {
         g_clear_pointer (&operation->main_thread_source,
                          g_source_destroy);
-        PK11_FreeSlot (operation->card_slot);
+        g_clear_object (&operation->card_slot);
+        g_clear_pointer (&operation->token_info, gck_token_info_free);
         g_free (operation->object_path);
         g_free (operation);
 }
@@ -617,11 +632,11 @@ static gboolean
 on_main_thread_to_register_new_token (GTask *task)
 {
         GsdSmartcardService *self;
-        GDBusObjectSkeleton *object;
-        GDBusInterfaceSkeleton *interface;
         RegisterNewTokenOperation *operation;
-        SECMODModule *driver;
-        char *driver_object_path;
+        g_autoptr(GDBusObjectSkeleton) object = NULL;
+        g_autoptr(GDBusInterfaceSkeleton) interface = NULL;
+        g_autoptr(GckModule) module = NULL;
+        g_autofree char *driver_object_path = NULL;
         const char *token_name;
 
         self = g_task_get_source_object (task);
@@ -633,18 +648,15 @@ on_main_thread_to_register_new_token (GTask *task)
         interface = G_DBUS_INTERFACE_SKELETON (gsd_smartcard_service_token_skeleton_new ());
 
         g_dbus_object_skeleton_add_interface (object, interface);
-        g_object_unref (interface);
 
-        driver = PK11_GetModule (operation->card_slot);
-        driver_object_path = get_object_path_for_driver (self, driver);
-
-        token_name = PK11_GetTokenName (operation->card_slot);
+        module = gck_slot_get_module (operation->card_slot);
+        driver_object_path = get_object_path_for_module (self, module);
+        token_name = operation->token_info->label;
 
         g_object_set (G_OBJECT (interface),
                       "driver", driver_object_path,
                       "name", token_name,
                       NULL);
-        g_free (driver_object_path);
 
         g_dbus_object_manager_server_export (self->object_manager_server,
                                              object);
@@ -664,19 +676,18 @@ create_main_thread_source (GSourceFunc   callback,
                            gpointer      user_data,
                            GSource     **source_out)
 {
-        GSource *source;
+        g_autoptr(GSource) source = NULL;
 
         source = g_idle_source_new ();
         g_source_set_callback (source, callback, user_data, NULL);
 
         *source_out = source;
         g_source_attach (source, NULL);
-        g_source_unref (source);
 }
 
 static void
 register_new_token_in_main_thread (GsdSmartcardService *self,
-                                   PK11SlotInfo        *card_slot,
+                                   GckSlot             *card_slot,
                                    char                *object_path,
                                    GCancellable        *cancellable,
                                    GAsyncReadyCallback  callback,
@@ -686,7 +697,8 @@ register_new_token_in_main_thread (GsdSmartcardService *self,
         GTask *task;
 
         operation = g_new0 (RegisterNewTokenOperation, 1);
-        operation->card_slot = PK11_ReferenceSlot (card_slot);
+        operation->card_slot = g_object_ref (card_slot);
+        operation->token_info = gck_slot_get_token_info (card_slot);
         operation->object_path = g_strdup (object_path);
 
         task = g_task_new (self, cancellable, callback, user_data);
@@ -711,8 +723,9 @@ register_new_token_in_main_thread_finish (GsdSmartcardService  *self,
 static void
 on_token_registered (GsdSmartcardService *self,
                      GAsyncResult        *result,
-                     PK11SlotInfo        *card_slot)
+                     gpointer             user_data)
 {
+        g_autoptr(GckSlot) card_slot = g_steal_pointer (&user_data);
         gboolean registered;
         GError *error = NULL;
 
@@ -721,19 +734,16 @@ on_token_registered (GsdSmartcardService *self,
         if (!registered) {
                 g_debug ("Couldn't register token: %s",
                          error->message);
-                goto out;
+                return;
         }
 
         synchronize_token_now (self, card_slot);
-
-out:
-        PK11_FreeSlot (card_slot);
 }
 
 typedef struct
 {
-        PK11SlotInfo *card_slot;
-        GSource      *main_thread_source;
+        GckSlot *card_slot;
+        GSource *main_thread_source;
 } SynchronizeTokenOperation;
 
 static void
@@ -741,7 +751,7 @@ destroy_synchronize_token_operation (SynchronizeTokenOperation *operation)
 {
         g_clear_pointer (&operation->main_thread_source,
                          g_source_destroy);
-        PK11_FreeSlot (operation->card_slot);
+        g_clear_object (&operation->card_slot);
         g_free (operation);
 }
 
@@ -774,7 +784,7 @@ synchronize_token_in_main_thread_finish (GsdSmartcardService  *self,
 
 static void
 synchronize_token_in_main_thread (GsdSmartcardService *self,
-                                  PK11SlotInfo        *card_slot,
+                                  GckSlot             *card_slot,
                                   GCancellable        *cancellable,
                                   GAsyncReadyCallback  callback,
                                   gpointer             user_data)
@@ -783,7 +793,7 @@ synchronize_token_in_main_thread (GsdSmartcardService *self,
         GTask *task;
 
         operation = g_new0 (SynchronizeTokenOperation, 1);
-        operation->card_slot = PK11_ReferenceSlot (card_slot);
+        operation->card_slot = g_object_ref (card_slot);
 
         task = g_task_new (self, cancellable, callback, user_data);
 
@@ -801,7 +811,7 @@ synchronize_token_in_main_thread (GsdSmartcardService *self,
 static void
 on_token_synchronized (GsdSmartcardService *self,
                        GAsyncResult        *result,
-                       PK11SlotInfo        *card_slot)
+                       gpointer             user_data)
 {
         gboolean synchronized;
         GError *error = NULL;
@@ -810,16 +820,14 @@ on_token_synchronized (GsdSmartcardService *self,
 
         if (!synchronized)
                 g_debug ("Couldn't synchronize token: %s", error->message);
-
-        PK11_FreeSlot (card_slot);
 }
 
 void
 gsd_smartcard_service_sync_token (GsdSmartcardService *self,
-                                  PK11SlotInfo        *card_slot,
+                                  GckSlot             *card_slot,
                                   GCancellable        *cancellable)
 {
-        char *object_path;
+        g_autofree char *object_path = NULL;
         GDBusInterfaceSkeleton *interface;
 
         object_path = get_object_path_for_token (self, card_slot);
@@ -835,7 +843,7 @@ gsd_smartcard_service_sync_token (GsdSmartcardService *self,
                                                    cancellable,
                                                    (GAsyncReadyCallback)
                                                    on_token_registered,
-                                                   PK11_ReferenceSlot (card_slot));
+                                                   g_object_ref (card_slot));
 
         else
                 synchronize_token_in_main_thread (self,
@@ -843,7 +851,5 @@ gsd_smartcard_service_sync_token (GsdSmartcardService *self,
                                                   cancellable,
                                                   (GAsyncReadyCallback)
                                                   on_token_synchronized,
-                                                  PK11_ReferenceSlot (card_slot));
-
-        g_free (object_path);
+                                                  NULL);
 }
